urcu: use unsigned long instead of long
[urcu.git] / urcu.c
diff --git a/urcu.c b/urcu.c
index 67a3eb21c8996b0751f9fc1d14ed9d71379f8bf5..140cd19b98add54f276755e08b13bfad9e8c3d0c 100644 (file)
--- a/urcu.c
+++ b/urcu.c
@@ -57,7 +57,7 @@ void __attribute__((constructor)) rcu_init(void);
 void __attribute__((destructor)) rcu_exit(void);
 #endif
 
-static pthread_mutex_t rcu_mutex = PTHREAD_MUTEX_INITIALIZER;
+static pthread_mutex_t rcu_gp_lock = PTHREAD_MUTEX_INITIALIZER;
 
 int gp_futex;
 
@@ -67,7 +67,7 @@ int gp_futex;
  * Also has a RCU_GP_COUNT of 1, to accelerate the reader fast path.
  * Written to only by writer with mutex taken. Read by both writer and readers.
  */
-long rcu_gp_ctr = RCU_GP_COUNT;
+unsigned long rcu_gp_ctr = RCU_GP_COUNT;
 
 /*
  * Written to only by each individual reader. Read by both the reader and the
@@ -82,18 +82,18 @@ unsigned int __thread rand_yield;
 
 static LIST_HEAD(registry);
 
-static void internal_rcu_lock(void)
+static void mutex_lock(pthread_mutex_t *mutex)
 {
        int ret;
 
 #ifndef DISTRUST_SIGNALS_EXTREME
-       ret = pthread_mutex_lock(&rcu_mutex);
+       ret = pthread_mutex_lock(mutex);
        if (ret) {
                perror("Error in pthread mutex lock");
                exit(-1);
        }
 #else /* #ifndef DISTRUST_SIGNALS_EXTREME */
-       while ((ret = pthread_mutex_trylock(&rcu_mutex)) != 0) {
+       while ((ret = pthread_mutex_trylock(mutex)) != 0) {
                if (ret != EBUSY && ret != EINTR) {
                        printf("ret = %d, errno = %d\n", ret, errno);
                        perror("Error in pthread mutex lock");
@@ -109,37 +109,29 @@ static void internal_rcu_lock(void)
 #endif /* #else #ifndef DISTRUST_SIGNALS_EXTREME */
 }
 
-static void internal_rcu_unlock(void)
+static void mutex_unlock(pthread_mutex_t *mutex)
 {
        int ret;
 
-       ret = pthread_mutex_unlock(&rcu_mutex);
+       ret = pthread_mutex_unlock(mutex);
        if (ret) {
                perror("Error in pthread mutex unlock");
                exit(-1);
        }
 }
 
-/*
- * called with rcu_mutex held.
- */
-static void switch_next_rcu_qparity(void)
-{
-       STORE_SHARED(rcu_gp_ctr, rcu_gp_ctr ^ RCU_GP_CTR_PHASE);
-}
-
 #ifdef RCU_MEMBARRIER
-static void smp_mb_heavy(void)
+static void smp_mb_master(int group)
 {
        if (likely(has_sys_membarrier))
-               membarrier(1);
+               membarrier(MEMBARRIER_EXPEDITED);
        else
                smp_mb();
 }
 #endif
 
 #ifdef RCU_MB
-static void smp_mb_heavy(void)
+static void smp_mb_master(int group)
 {
        smp_mb();
 }
@@ -189,7 +181,7 @@ static void force_mb_all_readers(void)
        smp_mb();       /* read ->need_mb before ending the barrier */
 }
 
-static void smp_mb_heavy(void)
+static void smp_mb_master(int group)
 {
        force_mb_all_readers();
 }
@@ -201,20 +193,35 @@ static void smp_mb_heavy(void)
 static void wait_gp(void)
 {
        /* Read reader_gp before read futex */
-       smp_mb_heavy();
+       smp_mb_master(RCU_MB_GROUP);
        if (uatomic_read(&gp_futex) == -1)
                futex_async(&gp_futex, FUTEX_WAIT, -1,
                      NULL, NULL, 0);
 }
 
-void wait_for_quiescent_state(void)
+void update_counter_and_wait(void)
 {
        LIST_HEAD(qsreaders);
        int wait_loops = 0;
        struct rcu_reader *index, *tmp;
 
-       if (list_empty(&registry))
-               return;
+       /* Switch parity: 0 -> 1, 1 -> 0 */
+       STORE_SHARED(rcu_gp_ctr, rcu_gp_ctr ^ RCU_GP_CTR_PHASE);
+
+       /*
+        * Must commit qparity update to memory before waiting for other parity
+        * quiescent state. Failure to do so could result in the writer waiting
+        * forever while new readers are always accessing data (no progress).
+        * Ensured by STORE_SHARED and LOAD_SHARED.
+        */
+
+       /*
+        * Adding a smp_mb() which is _not_ formally required, but makes the
+        * model easier to understand. It does not have a big performance impact
+        * anyway, given this is the write-side.
+        */
+       smp_mb();
+
        /*
         * Wait for each thread rcu_reader.ctr count to become 0.
         */
@@ -223,7 +230,7 @@ void wait_for_quiescent_state(void)
                if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
                        uatomic_dec(&gp_futex);
                        /* Write futex before read reader_gp */
-                       smp_mb_heavy();
+                       smp_mb_master(RCU_MB_GROUP);
                }
 
                list_for_each_entry_safe(index, tmp, &registry, head) {
@@ -235,7 +242,7 @@ void wait_for_quiescent_state(void)
                if (list_empty(&registry)) {
                        if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
-                               smp_mb_heavy();
+                               smp_mb_master(RCU_MB_GROUP);
                                uatomic_set(&gp_futex, 0);
                        }
                        break;
@@ -253,7 +260,7 @@ void wait_for_quiescent_state(void)
                if (list_empty(&registry)) {
                        if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS) {
                                /* Read reader_gp before write futex */
-                               smp_mb_heavy();
+                               smp_mb_master(RCU_MB_GROUP);
                                uatomic_set(&gp_futex, 0);
                        }
                        break;
@@ -263,7 +270,7 @@ void wait_for_quiescent_state(void)
                                wait_gp();
                                break; /* only escape switch */
                        case KICK_READER_LOOPS:
-                               smp_mb_heavy();
+                               smp_mb_master(RCU_MB_GROUP);
                                wait_loops = 0;
                                break; /* only escape switch */
                        default:
@@ -278,35 +285,21 @@ void wait_for_quiescent_state(void)
 
 void synchronize_rcu(void)
 {
-       internal_rcu_lock();
+       mutex_lock(&rcu_gp_lock);
+
+       if (list_empty(&registry))
+               goto out;
 
        /* All threads should read qparity before accessing data structure
* where new ptr points to. Must be done within internal_rcu_lock
* because it iterates on reader threads.*/
        /* Write new ptr before changing the qparity */
-       smp_mb_heavy();
-
-       switch_next_rcu_qparity();      /* 0 -> 1 */
-
-       /*
-        * Must commit qparity update to memory before waiting for parity
-        * 0 quiescent state. Failure to do so could result in the writer
-        * waiting forever while new readers are always accessing data (no
-        * progress).
-        * Ensured by STORE_SHARED and LOAD_SHARED.
-        */
-
-       /*
-        * Adding a smp_mb() which is _not_ formally required, but makes the
-        * model easier to understand. It does not have a big performance impact
-        * anyway, given this is the write-side.
-        */
-       smp_mb();
+       smp_mb_master(RCU_MB_GROUP);
 
        /*
         * Wait for previous parity to be empty of readers.
         */
-       wait_for_quiescent_state();     /* Wait readers in parity 0 */
+       update_counter_and_wait();      /* 0 -> 1, wait readers in parity 0 */
 
        /*
         * Must finish waiting for quiescent state for parity 0 before
@@ -323,34 +316,17 @@ void synchronize_rcu(void)
         */
        smp_mb();
 
-       switch_next_rcu_qparity();      /* 1 -> 0 */
-
-       /*
-        * Must commit qparity update to memory before waiting for parity
-        * 1 quiescent state. Failure to do so could result in the writer
-        * waiting forever while new readers are always accessing data (no
-        * progress).
-        * Ensured by STORE_SHARED and LOAD_SHARED.
-        */
-
-       /*
-        * Adding a smp_mb() which is _not_ formally required, but makes the
-        * model easier to understand. It does not have a big performance impact
-        * anyway, given this is the write-side.
-        */
-       smp_mb();
-
        /*
         * Wait for previous parity to be empty of readers.
         */
-       wait_for_quiescent_state();     /* Wait readers in parity 1 */
+       update_counter_and_wait();      /* 1 -> 0, wait readers in parity 1 */
 
        /* Finish waiting for reader threads before letting the old ptr being
-        * freed. Must be done within internal_rcu_lock because it iterates on
-        * reader threads. */
-       smp_mb_heavy();
-
-       internal_rcu_unlock();
+        * freed. Must be done within rcu_gp_lock because it iterates on reader
+        * threads. */
+       smp_mb_master(RCU_MB_GROUP);
+out:
+       mutex_unlock(&rcu_gp_lock);
 }
 
 /*
@@ -373,17 +349,17 @@ void rcu_register_thread(void)
        assert(rcu_reader.need_mb == 0);
        assert(rcu_reader.ctr == 0);
 
-       internal_rcu_lock();
+       mutex_lock(&rcu_gp_lock);
        rcu_init();     /* In case gcc does not support constructor attribute */
        list_add(&rcu_reader.head, &registry);
-       internal_rcu_unlock();
+       mutex_unlock(&rcu_gp_lock);
 }
 
 void rcu_unregister_thread(void)
 {
-       internal_rcu_lock();
+       mutex_lock(&rcu_gp_lock);
        list_del(&rcu_reader.head);
-       internal_rcu_unlock();
+       mutex_unlock(&rcu_gp_lock);
 }
 
 #ifdef RCU_MEMBARRIER
@@ -392,7 +368,7 @@ void rcu_init(void)
        if (init_done)
                return;
        init_done = 1;
-       if (!membarrier(1))
+       if (!membarrier(MEMBARRIER_EXPEDITED | MEMBARRIER_QUERY))
                has_sys_membarrier = 1;
 }
 #endif
@@ -414,9 +390,9 @@ static void sigrcu_handler(int signo, siginfo_t *siginfo, void *context)
  * rcu_init constructor. Called when the library is linked, but also when
  * reader threads are calling rcu_register_thread().
  * Should only be called by a single thread at a given time. This is ensured by
* holing the internal_rcu_lock() from rcu_register_thread() or by running at
* library load time, which should not be executed by multiple threads nor
* concurrently with rcu_register_thread() anyway.
  */
 void rcu_init(void)
 {
@@ -278,35 +285,21 @@ void wait_for_quiescent_state(void)
 
        /* All threads should read qparity before accessing data structure
-        * where new ptr points to. Must be done within internal_rcu_lock
-        * because it iterates on reader threads.*/
+        * where new ptr points to. Must be done within rcu_gp_lock because it
+        * iterates on reader threads.*/
        /* Write new ptr before changing the qparity */
@@ -414,9 +390,9 @@ static void sigrcu_handler(int signo, siginfo_t *siginfo, void *context)
  * reader threads are calling rcu_register_thread().
  * Should only be called by a single thread at a given time. This is ensured by
- * holing the internal_rcu_lock() from rcu_register_thread() or by running at
- * library load time, which should not be executed by multiple threads nor
- * concurrently with rcu_register_thread() anyway.
+ * holing the rcu_gp_lock from rcu_register_thread() or by running at library
+ * load time, which should not be executed by multiple threads nor concurrently
+ * with rcu_register_thread() anyway.
  */
 void rcu_init(void)
This page took 0.02629 seconds and 4 git commands to generate.